from typing import List
import ast

import logging

from ..module import Module, SafeFilenameModule
from .path import ImportPath


logger = logging.getLogger(__name__)


class DependencyAnalyzer:
    """
    Analyzes a set of Python modules for imports between them.

    Args:
        modules: list of all SafeFilenameModules that make up the package.
        package: the Python package that contains all the modules.

    Usage:
        analyzer = DependencyAnalyzer(modules)
        import_paths = analyzer.determine_import_paths()
    """
    def __init__(self, modules: List[SafeFilenameModule], package: Module) -> None:
        self.modules = modules
        self.package = package

    def determine_import_paths(self) -> List[ImportPath]:
        """
        Return a list of the ImportPaths for all the modules.
        """
        import_paths: List[ImportPath] = []
        for module in self.modules:
            import_paths.extend(
                self._determine_import_paths_for_module(module)
            )
        return import_paths

    def _determine_import_paths_for_module(self, module: SafeFilenameModule) -> List[ImportPath]:
        """
        Return a list of all the ImportPaths for which a given Module is the importer.
        """
        import_paths: List[ImportPath] = []
        imported_modules = self._get_imported_modules(module)
        for imported_module in imported_modules:
            import_paths.append(
                ImportPath(
                    importer=module,
                    imported=imported_module
                )
            )
        return import_paths

    def _get_imported_modules(self, module: SafeFilenameModule) -> List[Module]:
        """
        Statically analyses the given module and returns a list of Modules that it imports.

        Note: this method only analyses the module in question and will not load any other code,
        so it relies on self.modules to deduce which modules it imports. (This is because you
        can't know whether "from foo.bar import baz" is importing a module called `baz`,
        or a function `baz` from the module `bar`.)
        """
        imported_modules = []

        with open(module.filename) as file:
            module_contents = file.read()

        ast_tree = ast.parse(module_contents)
        for node in ast.walk(ast_tree):
            if isinstance(node, ast.ImportFrom):
                # Parsing something in the form 'from x import ...'.
                assert isinstance(node.level, int)
                if node.level == 0:
                    # Absolute import.
                    # Let the type checker know we expect node.module to be set here.
                    assert isinstance(node.module, str)
                    if not node.module.startswith(self.package.name):
                        # Don't include imports of modules outside this package.
                        continue
                    module_base = node.module
                elif node.level >= 1:
                    # Relative import. The level corresponds to how high up the tree it goes;
                    # for example 'from ... import foo' would be level 3.
                    importing_module_components = module.name.split('.')
                    # TODO: handle level that is too high.
                    # Trim the base module by the number of levels.
                    module_base = '.'.join(importing_module_components[:-node.level])
                    if node.module:
                        module_base = '.'.join([module_base, node.module])

                # node.names corresponds to 'a', 'b' and 'c' in 'from x import a, b, c'.
                for alias in node.names:
                    full_module_name = '.'.join([module_base, alias.name])
                    imported_modules.append(Module(full_module_name))

            elif isinstance(node, ast.Import):
                # Parsing a line in the form 'import x'.
                for alias in node.names:
                    if not alias.name.startswith(self.package.name):
                        # Don't include imports of modules outside this package.
                        continue
                    imported_modules.append(Module(alias.name))
            else:
                # Not an import statement; move on.
                continue

        imported_modules = self._trim_each_to_known_modules(imported_modules)
        return imported_modules

    def _trim_each_to_known_modules(self, imported_modules: List[Module]) -> List[Module]:
        known_modules = []
        for imported_module in imported_modules:
            if imported_module in self.modules:
                known_modules.append(imported_module)
            else:
                # The module isn't in the known modules. This is because it's something *within*
                # a module (e.g. a function): the result of something like 'from .subpackage
                # import my_function'. So we trim the components back to the module.
                components = imported_module.name.split('.')[:-1]
                trimmed_module = Module('.'.join(components))
                if trimmed_module in self.modules:
                    known_modules.append(trimmed_module)
                else:
                    # TODO: we may want to warn the user about this.
                    logger.debug('{} not found in modules.'.format(trimmed_module))
        return known_modules



from layer_linter.dependencies import DependencyGraph
from layer_linter.module import Module, SafeFilenameModule
import os
import sys


def test_dependency_graph():
    dirname = os.path.dirname(__file__)
    path = os.path.abspath(os.path.join(dirname, '..', '..', 'assets'))

    sys.path.append(path)

    ROOT_PACKAGE = 'dependenciespackage'
    MODULE_ONE = Module("{}.one".format(ROOT_PACKAGE))
    MODULE_TWO = Module("{}.two".format(ROOT_PACKAGE))
    MODULE_THREE = Module("{}.three".format(ROOT_PACKAGE))
    MODULE_FOUR = Module("{}.four".format(ROOT_PACKAGE))

    SUBPACKAGE = 'subpackage'
    SUBMODULE_ONE = Module("{}.{}.one".format(ROOT_PACKAGE, SUBPACKAGE))
    SUBMODULE_TWO = Module("{}.{}.two".format(ROOT_PACKAGE, SUBPACKAGE))
    SUBMODULE_THREE = Module("{}.{}.three".format(ROOT_PACKAGE, SUBPACKAGE))

    SUBSUBPACKAGE = 'subsubpackage'
    SUBSUBMODULE_ONE = Module("{}.{}.{}.one".format(ROOT_PACKAGE, SUBPACKAGE, SUBSUBPACKAGE))
    SUBSUBMODULE_TWO = Module("{}.{}.{}.two".format(ROOT_PACKAGE, SUBPACKAGE, SUBSUBPACKAGE))
    SUBSUBMODULE_THREE = Module("{}.{}.{}.three".format(ROOT_PACKAGE, SUBPACKAGE, SUBSUBPACKAGE))

    root_package = __import__(ROOT_PACKAGE)
    graph = DependencyGraph(
        SafeFilenameModule(name=ROOT_PACKAGE, filename=root_package.__file__)
    )

    assert graph.find_path(
        upstream=MODULE_ONE,
        downstream=MODULE_TWO) == (MODULE_TWO, MODULE_ONE)

    assert graph.find_path(
        upstream=MODULE_TWO,
        downstream=MODULE_ONE) is None

    assert graph.find_path(
        upstream=MODULE_ONE,
        downstream=MODULE_FOUR) == (MODULE_FOUR, MODULE_THREE,
                                    MODULE_TWO, MODULE_ONE)

    assert graph.find_path(
        upstream=SUBMODULE_ONE,
        downstream=SUBMODULE_THREE) == (SUBMODULE_THREE, SUBMODULE_TWO, SUBMODULE_ONE)

    assert graph.find_path(
        upstream=SUBSUBMODULE_ONE,
        downstream=SUBSUBMODULE_THREE) == (SUBSUBMODULE_THREE, SUBSUBMODULE_TWO, SUBSUBMODULE_ONE)

    # Module count should be 13 (running total in square brackets):
    # - dependenciespackage [1]
    #   - one [2]
    #   - two [3]
    #   - three [4]
    #   - four [5]
    #   - .hidden [X]
    #   - migrations [X]
    #   - subpackage [6]
    #     - one [7]
    #     - two [8]
    #     - three [9]
    #     - is [10] (treat reserved keywords as normal modules)
    #     - subsubpackage [11]
    #       - one [12]
    #       - two [13]
    #       - three [14]
    assert graph.module_count == 14
    # Dependency count should be 7:
    # dependenciespackage.two <- dependenciespackage.one
    # dependenciespackage.three <- dependenciespackage.two
    # dependenciespackage.four <- dependenciespackage.three
    # dependenciespackage.subpackage.two <- dependenciespackage.subpackage.one
    # dependenciespackage.subpackage.three <- dependenciespackage.subpackage.two
    # dependenciespackage.subpackage.is <- dependenciespackage.subpackage.three
    # dependenciespackage.subpackage.subsubpackage.two <- dependenciespackage.subpackage.subsubpackage.one
    # dependenciespackage.subpackage.subsubpackage.three <- dependenciespackage.subpackage.subsubpackage.two
    assert graph.dependency_count == 8



import os
from typing import Tuple, List

import pytest

from layer_linter.dependencies.scanner import PackageScanner
from layer_linter.module import Module, SafeFilenameModule


class TestPackageScanner:
    def test_success(self):
        package = SafeFilenameModule(
            name='scannersuccess',
            filename=os.path.join(self._get_package_directory('scannersuccess'), '__init__.py'),
        )

        scanner = PackageScanner(package)
        modules = scanner.scan_for_modules()

        expected_modules = self._build_modules(
            package_name='scannersuccess',
            tuples=(
                ('scannersuccess', '__init__.py'),
                ('scannersuccess.one', 'one/__init__.py'),
                ('scannersuccess.one.alpha', 'one/alpha.py'),
                ('scannersuccess.one.beta', 'one/beta.py'),
                ('scannersuccess.one.gamma', 'one/gamma.py'),
                ('scannersuccess.one.delta', 'one/delta/__init__.py'),
                ('scannersuccess.one.delta.green', 'one/delta/green.py'),
                ('scannersuccess.one.delta.red_blue', 'one/delta/red_blue.py'),
                ('scannersuccess.two', 'two/__init__.py'),
                ('scannersuccess.two.alpha', 'two/alpha.py'),
                ('scannersuccess.two.beta',  'two/beta.py'),
                ('scannersuccess.two.gamma',  'two/gamma.py'),
                ('scannersuccess.four', 'four.py'),
                # Also include invalid modules. To keep things simple, Layer Linter won't decide
                # what is and isn't allowed - but we don't want them to break things.
                ('scannersuccess.in', 'in/__init__.py'),
                ('scannersuccess.in.class', 'in/class.py'),
                ('scannersuccess.in.hyphenated-name', 'in/hyphenated-name.py'),

            ),
        )
        assert set(modules) == set(expected_modules)
        # Also check that the filenames are the same.
        sort_by_name = lambda m: m.name
        sorted_modules = sorted(modules, key=sort_by_name)
        sorted_expected_modules = sorted(expected_modules, key=sort_by_name)
        for index, module in enumerate(sorted_modules):
            assert module.filename == sorted_expected_modules[index].filename

    def _get_package_directory(self, package_name: str) -> str:
        """
        Return absolute path to the package directory.
        """
        return os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                '..', '..', 'assets', package_name, package_name,
            )
        )

    def _build_modules(self,
                       package_name: str,
                       tuples: Tuple[Tuple[str, str]]) -> List[SafeFilenameModule]:
        package_directory = self._get_package_directory(package_name)
        modules = []
        for module_name, module_filename in tuples:
            modules.append(
                SafeFilenameModule(
                    name=module_name,
                    filename=os.path.join(package_directory, module_filename),
                )
            )
        return modules



from typing import Any


class Module:
    """
    A Python module.
    """
    def __init__(self, name: str) -> None:
        """
        Args:
            name: The fully qualified name of a Python module, e.g. 'package.foo.bar'.
        """
        self.name = name

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return "<{}: {}>".format(self.__class__.__name__, self)

    def __eq__(self, other: Any) -> bool:
        # Only other Module instances with the same name are equal to each other.
        if isinstance(other, Module):
            return self.name == other.name
        else:
            return False

    def __hash__(self) -> int:
        return hash(str(self))


class SafeFilenameModule(Module):
    """
    A Python module whose filename can be known safely, without importing the code.
    """
    def __init__(self, name: str, filename: str) -> None:
        """
        Args:
            name: The fully qualified name of a Python module, e.g. 'package.foo.bar'.
            filename: The full filename and path to the Python file,
            e.g. '/path/to/package/one.py'.
        """
        self.filename = filename
        super().__init__(name)



from typing import List, Dict, Iterable, Optional
import re
import yaml
import importlib
import logging
from copy import copy

from .dependencies import DependencyGraph, ImportPath
from .module import Module


logger = logging.getLogger(__name__)


class ContractParseError(IOError):
    pass


class Layer:
    """
    A relatively named module or subpackage. When combined with a container,
    it refers to a specific module. For example, Layer('one') would refer
    to Module('foo.one') when combined with the containing Module('foo').

    If a Layer is marked as optional, then a Contract won't complain if
    it can't find it within a particular container.
    """
    def __init__(self, name: str, is_optional=False) -> None:
        self.name = name
        self.is_optional = is_optional

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return '<{}: {}>'.format(self.__class__.__name__, self)


class Contract:
    def __init__(self, name: str, containers: List[Module], layers: List[Layer],
                 whitelisted_paths: Optional[List[ImportPath]] = None) -> None:
        self.name = name
        self.containers = containers
        self.layers = layers
        self.whitelisted_paths = whitelisted_paths if whitelisted_paths else []

    def check_dependencies(self, dependencies: DependencyGraph) -> None:
        self._check_all_layers_exist_for_all_containers(dependencies)

        self.illegal_dependencies: List[List[str]] = []

        logger.debug('Checking dependencies for contract {}...'.format(self))

        for container in self.containers:
            for layer in reversed(self.layers):
                self._check_layer_does_not_import_downstream(layer, container, dependencies)

    def _check_all_layers_exist_for_all_containers(self, dependencies: DependencyGraph) -> None:
        """
        Raise a ValueError if we couldn't find any Python files for each layer in all containers.
        """
        for container in self.containers:
            self._check_all_layers_exist_for_container(container, dependencies)

    def _check_all_layers_exist_for_container(self,
                                              container: Module,
                                              dependencies: DependencyGraph) -> None:
        """
        Raise a ValueError if we couldn't find any Python files for each non-optional layer.
        """
        for layer in self.layers:
            if layer.is_optional:
                continue
            layer_module = self._get_layer_module(layer, container)
            if layer_module not in dependencies:
                raise ValueError(f"Missing layer in container '{container}': "
                                 f"module {layer_module} does not exist.")

    def _check_layer_does_not_import_downstream(self, layer: Layer, container: Module,
                                                dependencies: DependencyGraph) -> None:

        logger.debug("Layer '{}' in container '{}'.".format(layer, container))

        modules_in_this_layer = self._get_modules_in_layer(layer, container, dependencies)
        modules_in_downstream_layers = self._get_modules_in_downstream_layers(
            layer, container, dependencies)
        logger.debug('Modules in this layer: {}'.format(modules_in_this_layer))
        logger.debug('Modules in downstream layer: {}'.format(modules_in_downstream_layers))

        for upstream_module in modules_in_this_layer:
            for downstream_module in modules_in_downstream_layers:
                logger.debug('Upstream {}, downstream {}.'.format(upstream_module,
                                                                  downstream_module))
                path = dependencies.find_path(
                    upstream=downstream_module,
                    downstream=upstream_module,
                    ignore_paths=self.whitelisted_paths,
                )
                logger.debug('Path is {}.'.format(path))
                if path and not self._path_is_via_another_layer(path, layer, container):
                    logger.debug('Illegal dependency found: {}'.format(path))
                    self._update_illegal_dependencies(path)

    def _get_modules_in_layer(
        self, layer: Layer, container: Module, dependencies: DependencyGraph
    ) -> List[Module]:
        """
        Args:
            layer: The Layer object.
            container: absolute name of the package containing the layer (string).
            dependencies: the DependencyGraph object.
        Returns:
            List of modules names within that layer, including the layer module itself.
            Includes grandchildren and deeper.
        """
        layer_module = self._get_layer_module(layer, container)
        modules = [layer_module]
        modules.extend(
            dependencies.get_descendants(layer_module)
        )
        return modules

    def _get_layer_module(self, layer: Layer, container: Module) -> Module:
        return Module("{}.{}".format(container, layer.name))

    def _get_modules_in_downstream_layers(
        self, layer: Layer, container: Module, dependencies: DependencyGraph
    ) -> List[Module]:
        modules = []
        for downstream_layer in self._get_layers_downstream_of(layer):
            modules.extend(
                self._get_modules_in_layer(layer=downstream_layer, container=container,
                                           dependencies=dependencies)
            )
        return modules

    def _path_is_via_another_layer(self, path, current_layer, container):
        other_layers = list(copy(self.layers))
        other_layers.remove(current_layer)

        layer_modules = ['{}.{}'.format(container, layer.name) for layer in other_layers]
        modules_via = path[1:-1]

        return any(path_module in layer_modules for path_module in modules_via)

    def _update_illegal_dependencies(self, path):
        # Don't duplicate path. So if the path is already present in another dependency,
        # don't add it. If another dependency is present in this path, replace it with this one.
        new_path_set = set(path)
        logger.debug('Updating illegal dependencies with {}.'.format(path))
        illegal_dependencies_copy = self.illegal_dependencies[:]
        paths_to_remove = []
        add_path = True
        for existing_path in illegal_dependencies_copy:
            existing_path_set = set(existing_path)
            logger.debug('Comparing new path with {}...'.format(existing_path))
            if new_path_set.issubset(existing_path_set):
                # Remove the existing_path, as the new path will be more succinct.
                logger.debug('Removing existing.')
                paths_to_remove.append(existing_path)
                add_path = True
            elif existing_path_set.issubset(new_path_set):
                # Don't add the new path, it's implied more succinctly with the existing path.
                logger.debug('Skipping new path.')
                add_path = False

        logger.debug('Paths to remove: {}'.format(paths_to_remove))
        self.illegal_dependencies = [
            i for i in self.illegal_dependencies if i not in paths_to_remove
        ]
        if add_path:
            self.illegal_dependencies.append(path)

    @property
    def is_kept(self) -> bool:
        try:
            return len(self.illegal_dependencies) == 0
        except AttributeError:
            raise RuntimeError(
                'Cannot check whether contract is kept '
                'until check_dependencies is called.'
            )

    def _get_layers_downstream_of(self, layer: Layer) -> Iterable[Layer]:
        return reversed(self.layers[:self.layers.index(layer)])

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return '<{}: {}>'.format(self.__class__.__name__, self)


PARENTHESES_REGEX = re.compile(r'^\(.*\)$')


def contract_from_yaml(key: str, data: Dict, package_name: str) -> Contract:
    layers: List[Layer] = []
    if 'layers' not in data:
        raise ContractParseError(f"'{key}' is missing a list of layers.")
    for layer_name in data['layers']:
        if PARENTHESES_REGEX.match(layer_name):
            layer = Layer(layer_name[1:-1], is_optional=True)
        else:
            layer = Layer(layer_name)
        layers.append(layer)

    containers: List[Module] = []
    if 'containers' not in data:
        error_message = f"'{key}' is missing a list of containers."
        # Help users who are using the older format.
        if 'packages' in data:
            error_message += " (Tip: try renaming 'packages' to 'containers'.)"
        raise ContractParseError(error_message)

    for container_name in data['containers']:
        _validate_container_name(container_name, package_name)
        containers.append(Module(container_name))

    whitelisted_paths: List[ImportPath] = []
    for whitelist_data in data.get('whitelisted_paths', []):
        try:
            importer, imported = map(Module, whitelist_data.split(' <- '))
        except ValueError:
            raise ValueError('Whitelisted paths must be in the format '
                             '"importer.module <- imported.module".')

        whitelisted_paths.append(ImportPath(importer, imported))

    return Contract(
        name=key,
        containers=containers,
        layers=layers,
        whitelisted_paths=whitelisted_paths,
    )


def get_contracts(filename: str, package_name: str) -> List[Contract]:
    """Read in any contracts from the given filename.
    """
    contracts = []

    with open(filename, 'r') as file:
        try:
            data_from_yaml = yaml.load(file)
        except Exception as e:
            logger.debug(e)
            raise ContractParseError('Could not parse {}.'.format(filename))
        for key, data in data_from_yaml.items():
            contracts.append(contract_from_yaml(key, data, package_name))

    return contracts


def _validate_container_name(container_name, package_name):
    """Raise a ValueError if the suppled container name is not a valid container for the supplied
    package name.
    """
    if not container_name.startswith(package_name):
        raise ValueError(
            f"Invalid container '{container_name}': containers must be either a "
            f"subpackage of '{package_name}', or '{package_name}' itself."
        )

    # Check that the container actually exists.
    if importlib.util.find_spec(container_name) is None:
        raise ValueError(
            f"Invalid container '{container_name}': no such package."
        )



from unittest.mock import MagicMock, patch, call, sentinel
import pytest

from layer_linter.module import Module
from layer_linter.contract import Contract
from layer_linter import report
from layer_linter.report import QuietReport, NormalReport, VerboseReport, ConsolePrinter


class StubDependencyGraph:
    def __init__(self):
        self.modules = [
            Module('foo.one'),
            Module('foo.two'),
            Module('foo.three'),
        ]
        self.module_count = len(self.modules)
        self.dependency_count = 4

    def get_modules_directly_imported_by(self, importer):
        return {
            self.modules[0]: [
                Module('foo.four.one'),
                Module('foo.four.two'),
            ],
            self.modules[1]: [],
            self.modules[2]: [
                Module('foo.five'),
            ]
        }[importer]


@pytest.fixture
def make_kept_contract():
    def _make_contract():
        contract = Contract('Foo contract', [], [])
        contract.check_dependencies(None)
        return contract
    return _make_contract


@pytest.fixture
def make_broken_contract():
    def _make_contract(name, illegal_dependencies=None):
        contract = Contract(name, [], [])
        if illegal_dependencies is None:
            illegal_dependencies = [
                ('foo.upstream', 'foo.downstream'),
            ]
        contract.illegal_dependencies = illegal_dependencies
        return contract
    return _make_contract


@pytest.mark.parametrize(
    'report_class', (
        QuietReport, NormalReport, VerboseReport,
    )
)
@patch.object(report, 'ConsolePrinter')
class TestReport:
    def _report_contracts(self, report_class, contracts):
        """
        Report on the supplied list of Contracts.
        """
        self.report = report_class(graph=StubDependencyGraph())

        for contract in contracts:
            self.report.add_contract(contract)
        self.report.output()

    def test_single_kept_contract(self, printer, report_class, make_kept_contract):
        kept_contract = make_kept_contract()

        self._report_contracts(report_class, [kept_contract])

        self.assert_headings_printed(report_class, printer, is_successful=True)

        if report_class is VerboseReport:
            self.assert_dependencies_printed(printer)

        if report_class is not QuietReport:
            printer.print_contract_one_liner.assert_called_once_with(kept_contract)
            printer.print_success.assert_called_once_with('Contracts: 1 kept, 0 broken.')

        assert self.report.has_broken_contracts is False

    def test_single_broken_contract_direct(self, printer, report_class, make_broken_contract):
        broken_contract = make_broken_contract(
            name='Foo',
            illegal_dependencies=[
                ('foo.two', 'foo.one'),
            ]
        )

        self._report_contracts(report_class, [broken_contract])

        self.assert_headings_printed(report_class, printer, is_successful=False)

        if report_class is VerboseReport:
            self.assert_dependencies_printed(printer)

        printer.print_error.assert_has_calls([call('Contracts: 0 kept, 1 broken.')])
        printer.print_heading.assert_has_calls([
            call('Broken contracts', printer.HEADING_LEVEL_TWO,
                 style=printer.ERROR),
            call('Foo', printer.HEADING_LEVEL_THREE,
                 style=printer.ERROR),
        ])
        printer.print_error.assert_has_calls([
            call('1. foo.two imports foo.one:'),
            call('foo.two <-', bold=False),
            call('foo.one', bold=False),
        ])

        assert self.report.has_broken_contracts is True

    def test_multiple_kept_and_broken_contracts(self, printer, report_class, make_kept_contract,
                                                make_broken_contract):
        broken_contract_1 = make_broken_contract(
            name='Contract 1',
            illegal_dependencies=[
                ('foo.two', 'foo.one'),
            ]
        )
        broken_contract_2 = make_broken_contract(
            name='Contract 2',
            illegal_dependencies=[
                ('bar.two', 'bar.one'),
            ]
        )
        kept_contract_1 = make_kept_contract()
        kept_contract_2 = make_kept_contract()
        kept_contract_3 = make_kept_contract()
        self._report_contracts(
            report_class,
            [
                broken_contract_1,
                kept_contract_1,
                broken_contract_2,
                kept_contract_2,
                kept_contract_3,
            ]
        )

        self.assert_headings_printed(report_class, printer, is_successful=False)

        if report_class is VerboseReport:
            self.assert_dependencies_printed(printer)

        printer.print_contract_one_liner.assert_has_calls([
            call(kept_contract_1),
            call(kept_contract_2),
            call(kept_contract_3),
            call(broken_contract_1),
            call(broken_contract_2),
        ])
        printer.print_error.assert_has_calls([call('Contracts: 3 kept, 2 broken.')])
        printer.print_heading.assert_has_calls([
            call('Broken contracts', printer.HEADING_LEVEL_TWO,
                 style=printer.ERROR),
            call('Contract 1', printer.HEADING_LEVEL_THREE,
                 style=printer.ERROR),
        ])
        printer.print_error.assert_has_calls([
            call('1. foo.two imports foo.one:'),
            call('foo.two <-', bold=False),
            call('foo.one', bold=False),
        ])
        printer.print_heading.assert_has_calls([
            call('Contract 2', printer.HEADING_LEVEL_THREE,
                 style=printer.ERROR),
        ])
        printer.print_error.assert_has_calls([
            call('1. bar.two imports bar.one:'),
            call('bar.two <-', bold=False),
            call('bar.one', bold=False),
        ])
        assert self.report.has_broken_contracts is True

    def test_single_broken_contract_indirect(self, printer, report_class, make_broken_contract):
        contract = make_broken_contract(
            name='Foo',
            illegal_dependencies=[
                ('foo.four', 'foo.three', 'foo.two', 'foo.one'),
            ]
        )

        self._report_contracts(report_class, [contract])

        self.assert_headings_printed(report_class, printer, is_successful=False)

        if report_class is VerboseReport:
            self.assert_dependencies_printed(printer)

        printer.print_contract_one_liner.assert_has_calls([
            call(contract),
        ])
        printer.print_error.assert_has_calls([call('Contracts: 0 kept, 1 broken.')])
        printer.print_heading.assert_has_calls([
            call('Broken contracts', printer.HEADING_LEVEL_TWO,
                 style=printer.ERROR),
            call('Foo', printer.HEADING_LEVEL_THREE,
                 style=printer.ERROR),
        ])
        printer.print_error.assert_has_calls([
            call('1. foo.four imports foo.one:'),
            call('foo.four <-', bold=False),
            call('foo.three <-', bold=False),
            call('foo.two <-', bold=False),
            call('foo.one', bold=False),
        ])
        assert self.report.has_broken_contracts is True

    def test_single_broken_contract_multiple_illegal_dependencies(self, printer, report_class,
                                                                  make_broken_contract):
        contract = make_broken_contract(
            name='Foo',
            illegal_dependencies=[
                ('foo.four', 'foo.three', 'foo.two', 'foo.one'),
                ('bar.two', 'bar.one'),
            ]
        )

        self._report_contracts(report_class, [contract])

        self.assert_headings_printed(report_class, printer, is_successful=False)

        if report_class is VerboseReport:
            self.assert_dependencies_printed(printer)
        
        printer.print_contract_one_liner.assert_has_calls([
            call(contract),
        ])
        printer.print_error.assert_has_calls([call('Contracts: 0 kept, 1 broken.')])
        printer.print_heading.assert_has_calls([
            call('Broken contracts', printer.HEADING_LEVEL_TWO,
                 style=printer.ERROR),
            call('Foo', printer.HEADING_LEVEL_THREE,
                 style=printer.ERROR),
        ])
        printer.print_error.assert_has_calls([
            call('1. foo.four imports foo.one:'),
            call('foo.four <-', bold=False),
            call('foo.three <-', bold=False),
            call('foo.two <-', bold=False),
            call('foo.one', bold=False),
            call('2. bar.two imports bar.one:'),
            call('bar.two <-', bold=False),
            call('bar.one', bold=False),
        ])
        assert self.report.has_broken_contracts is True

    def assert_headings_printed(self, report_class, printer, is_successful):
        if (report_class is QuietReport) and is_successful:
            printer.print_heading.assert_not_called()
            printer.print_contract_one_liner.assert_not_called()
            printer.print_success.assert_not_called()
        elif report_class is VerboseReport:
            printer.print_heading.assert_has_calls([
                call('Layer Linter', printer.HEADING_LEVEL_ONE),
                call('Dependencies', printer.HEADING_LEVEL_TWO),
                call('Contracts', printer.HEADING_LEVEL_TWO),
            ])
        else:
            printer.print_heading.assert_has_calls([
                call('Layer Linter', printer.HEADING_LEVEL_ONE),
                call('Contracts', printer.HEADING_LEVEL_TWO),
            ])

    def assert_dependencies_printed(self, printer):
        printer.print.assert_has_calls([
            call('foo.one imports:', bold=True),
            call('- foo.four.one'),
            call('- foo.four.two'),
            call('foo.two imports:', bold=True),
            call('- (nothing)'),
            call('foo.three imports:', bold=True),
            call('- foo.five'),
        ])


@patch.object(report, 'click')
class TestConsolePrinter:
    @pytest.mark.parametrize('style,expected_color', [
        (None, None),
        (ConsolePrinter.SUCCESS, 'green',),
        (ConsolePrinter.ERROR, 'red'),
    ])
    def test_print_heading_1(self, click, style, expected_color):
        ConsolePrinter.print_heading(
            'Lorem ipsum', ConsolePrinter.HEADING_LEVEL_ONE, style)

        assert click.secho.call_count == 3
        click.secho.assert_has_calls([
            call('===========', bold=True, fg=expected_color),
            call('Lorem ipsum', bold=True, fg=expected_color),
            call('===========', bold=True, fg=expected_color),
        ])
        click.echo.assert_called_once_with()

    def test_print_heading_2(self, click):
        ConsolePrinter.print_heading(
            'Dolor sic', ConsolePrinter.HEADING_LEVEL_TWO)

        assert click.secho.call_count == 3
        click.secho.assert_has_calls([
            call('---------', bold=True, fg=None),
            call('Dolor sic', bold=True, fg=None),
            call('---------', bold=True, fg=None),
        ])
        click.echo.assert_called_once_with()

    def test_print_heading_3(self, click):
        ConsolePrinter.print_heading(
            'Amet consectetur', ConsolePrinter.HEADING_LEVEL_THREE)

        assert click.secho.call_count == 2
        click.secho.assert_has_calls([
            call('Amet consectetur', bold=True, fg=None),
            call('----------------', bold=True, fg=None),
        ])
        click.echo.assert_called_once_with()

    @pytest.mark.parametrize('bold', (True, False))
    def test_print(self, click, bold):
        ConsolePrinter.print(sentinel.text, bold)

        click.secho.assert_called_once_with(
            sentinel.text, bold=bold)

    @pytest.mark.parametrize('bold', (True, False))
    def test_print_success(self, click, bold):
        ConsolePrinter.print_success(sentinel.text, bold)

        click.secho.assert_called_once_with(
            sentinel.text, fg='green', bold=bold)

    @pytest.mark.parametrize('bold', (True, False))
    def test_print_error(self, click, bold):
        ConsolePrinter.print_error(sentinel.text, bold)

        click.secho.assert_called_once_with(
            sentinel.text, fg='red', bold=bold)

    def test_indent_cursor(self, click):
        ConsolePrinter.indent_cursor()

        click.echo.assert_called_once_with('    ', nl=False)

    def test_new_line(self, click):
        ConsolePrinter.new_line()

        click.echo.assert_called_once_with()

    @pytest.mark.parametrize(
        'is_kept,whitelisted_path_length,expected_label,expected_color', [
            (True, 4, 'KEPT', 'green'),
            (False, 0, 'BROKEN', 'red'),
        ])
    def test_print_contract_one_liner(self, click, is_kept, whitelisted_path_length,
                                      expected_label, expected_color):
        contract = MagicMock(
            whitelisted_paths=[MagicMock()] * whitelisted_path_length,
            is_kept=is_kept,
        )
        contract.__str__.return_value = 'Foo'

        ConsolePrinter.print_contract_one_liner(contract)

        if whitelisted_path_length:
            click.secho.assert_has_calls([
                call('Foo ', nl=False),
                call('({} whitelisted paths) '.format(whitelisted_path_length), nl=False),
                call(expected_label, fg=expected_color, bold=True)
            ])
        else:
            click.secho.assert_has_calls([
                call('Foo ', nl=False),
                call(expected_label, fg=expected_color, bold=True)
            ])



from typing import List, Type

import click

from .dependencies import DependencyGraph
from .contract import Contract

VERBOSITY_QUIET = 0
VERBOSITY_NORMAL = 1
VERBOSITY_HIGH = 2


def get_report_class(verbosity: int) -> Type['BaseReport']:
    return {
        VERBOSITY_QUIET: QuietReport,
        VERBOSITY_NORMAL: NormalReport,
        VERBOSITY_HIGH: VerboseReport,
    }[verbosity]


class BaseReport:
    def __init__(self, graph: DependencyGraph) -> None:
        self.graph = graph
        self.kept_contracts: List[Contract] = []
        self.broken_contracts: List[Contract] = []
        self.has_broken_contracts = False

    def add_contract(self, contract: Contract) -> None:
        if contract.is_kept:
            self.kept_contracts.append(contract)
        else:
            self.broken_contracts.append(contract)
            self.has_broken_contracts = True

    def output(self) -> None:
        self._output_title()
        self._output_contracts_analysis()
        if self.broken_contracts:
            self._output_broken_contracts_details()

    def _output_title(self):
        ConsolePrinter.print_heading('Layer Linter', ConsolePrinter.HEADING_LEVEL_ONE)

    def _output_contracts_analysis(self):
        ConsolePrinter.print_heading('Contracts', ConsolePrinter.HEADING_LEVEL_TWO)

        ConsolePrinter.print_heading(
            'Analyzed {} files, {} dependencies.'.format(
                self.graph.module_count,
                self.graph.dependency_count,
            ), ConsolePrinter.HEADING_LEVEL_THREE)

        for contract in self.kept_contracts:
            self._output_contract_analysis(contract)

        for contract in self.broken_contracts:
            self._output_contract_analysis(contract)

        self._output_contracts_summary()

    def _output_contract_analysis(self, contract):
        ConsolePrinter.print_contract_one_liner(contract)

    def _output_contracts_summary(self):
        # Print summary line.
        ConsolePrinter.new_line()
        print_callback = ConsolePrinter.print_error if self.broken_contracts else \
            ConsolePrinter.print_success
        print_callback('Contracts: {} kept, {} broken.'.format(
            len(self.kept_contracts),
            len(self.broken_contracts)
        ))
        ConsolePrinter.new_line()

    def _output_broken_contracts_details(self):
        ConsolePrinter.print_heading('Broken contracts',
                                     ConsolePrinter.HEADING_LEVEL_TWO,
                                     style=ConsolePrinter.ERROR)
        ConsolePrinter.new_line()

        for broken_contract in self.broken_contracts:
            ConsolePrinter.print_heading(str(broken_contract), ConsolePrinter.HEADING_LEVEL_THREE,
                                         style=ConsolePrinter.ERROR)

            ConsolePrinter.new_line()

            for index, illegal_dependency in enumerate(broken_contract.illegal_dependencies):
                first = illegal_dependency[0]
                last = illegal_dependency[-1]

                ConsolePrinter.print_error('{}. {} imports {}:'.format(index + 1, first, last))
                ConsolePrinter.new_line()

                for dep_index, item in enumerate(illegal_dependency):
                    error_text = str(item)
                    if dep_index < len(illegal_dependency) - 1:
                        error_text += ' <-'
                    ConsolePrinter.indent_cursor()
                    ConsolePrinter.print_error(error_text, bold=False)
                ConsolePrinter.new_line()


class NormalReport(BaseReport):
    pass


class QuietReport(NormalReport):
    """
    Report that only reports when there is a broken contract.
    """
    def output(self) -> None:
        if self.broken_contracts:
            super().output()


class VerboseReport(BaseReport):
    def output(self):
        self._output_title()
        self._output_dependencies()
        self._output_contracts_analysis()
        if self.broken_contracts:
            self._output_broken_contracts_details()

    def _output_dependencies(self):
        ConsolePrinter.print_heading('Dependencies', ConsolePrinter.HEADING_LEVEL_TWO)

        for importer in self.graph.modules:
            ConsolePrinter.print('{} imports:'.format(importer), bold=True)
            has_at_least_one_successor = False
            for imported in self.graph.get_modules_directly_imported_by(importer=importer):
                has_at_least_one_successor = True
                ConsolePrinter.indent_cursor()
                ConsolePrinter.print('- {}'.format(imported))
            if not has_at_least_one_successor:
                ConsolePrinter.indent_cursor()
                ConsolePrinter.print('- (nothing)')
        ConsolePrinter.new_line()


class ConsolePrinter:
    ERROR = 'error'
    SUCCESS = 'success'
    COLORS = {
        ERROR: 'red',
        SUCCESS: 'green',
    }

    HEADING_LEVEL_ONE = 1
    HEADING_LEVEL_TWO = 2
    HEADING_LEVEL_THREE = 3

    HEADING_MAP = {
        HEADING_LEVEL_ONE: ('=', True),
        HEADING_LEVEL_TWO: ('-', True),
        HEADING_LEVEL_THREE: ('-', False),
    }

    INDENT_SIZE = 4

    @classmethod
    def print_heading(cls, text, level, style=None):
        """
        Prints the supplied text to the console, formatted as a heading.

        Args:
            text (str): the text to format as a heading.
            level (int): the level of heading to display (one of the keys of HEADING_MAP).
            style (str, optional): ERROR or SUCCESS style to apply (default None).
        Usage:

            ConsolePrinter.print_heading('Foo', ConsolePrinter.HEADING_LEVEL_ONE)
        """
        # Setup styling variables.
        is_bold = True
        color = cls.COLORS[style] if style else None
        line_char, show_line_above = cls.HEADING_MAP[level]
        heading_line = line_char * len(text)

        # Print lines.
        if show_line_above:
            click.secho(heading_line, bold=is_bold, fg=color)
        click.secho(text, bold=is_bold, fg=color)
        click.secho(heading_line, bold=is_bold, fg=color)
        click.echo()

    @classmethod
    def print(cls, text, bold=False):
        """
        Prints a line to the console.
        """
        click.secho(text, bold=bold)

    @classmethod
    def print_success(cls, text, bold=True):
        """
        Prints a line to the console, formatted as a success.
        """
        click.secho(text, fg=cls.COLORS[cls.SUCCESS], bold=bold)

    @classmethod
    def print_error(cls, text, bold=True):
        """
        Prints a line to the console, formatted as an error.
        """
        click.secho(text, fg=cls.COLORS[cls.ERROR], bold=bold)

    @classmethod
    def indent_cursor(cls):
        """
        Indents the cursor ready to print a line.
        """
        click.echo(' ' * cls.INDENT_SIZE, nl=False)

    @classmethod
    def new_line(cls):
        click.echo()

    @classmethod
    def print_contract_one_liner(cls, contract: Contract) -> None:
        is_kept = contract.is_kept
        click.secho('{} '.format(contract), nl=False)
        if contract.whitelisted_paths:
            click.secho('({} whitelisted paths) '.format(len(contract.whitelisted_paths)),
                        nl=False)
        status_map = {
            True: ('KEPT', cls.SUCCESS),
            False: ('BROKEN', cls.ERROR),
        }
        color = cls.COLORS[status_map[is_kept][1]]
        click.secho(status_map[is_kept][0], fg=color, bold=True)



from unittest.mock import patch

import pytest

from layer_linter import cmdline
from layer_linter.cmdline import _normalise_verbosity, _main
from layer_linter.report import VERBOSITY_HIGH, VERBOSITY_NORMAL, VERBOSITY_QUIET


@pytest.mark.parametrize(
    'verbosity,is_quiet,expected_result', (
        (0, True, VERBOSITY_QUIET),
        (0, False, VERBOSITY_NORMAL),
        (1, False, VERBOSITY_HIGH),
        (1, True, "Invalid parameters: quiet and verbose called together. Choose one or the "
                  "other."),
        (2, False, "That level of verbosity is not supported. Maximum verbosity is -v."),
    )
)
def test_normalise_verbosity(verbosity, is_quiet, expected_result):
    if expected_result in (VERBOSITY_QUIET, VERBOSITY_NORMAL, VERBOSITY_HIGH):
        result = _normalise_verbosity(verbosity, is_quiet)
        assert result == expected_result
    else:
        with pytest.raises(RuntimeError) as e:
            _normalise_verbosity(verbosity, is_quiet)
        assert str(e.value) == expected_result


@pytest.mark.parametrize(
    'is_debug', (True, False)
)
@patch.object(cmdline, 'get_report_class')
@patch.object(cmdline, 'DependencyGraph')
@patch.object(cmdline, '_get_package')
@patch.object(cmdline, '_get_contracts')
@patch.object(cmdline, 'logging')
def test_debug(mock_logging, mock_get_contracts, mock_get_package, mock_graph,
               mock_get_report_class, is_debug):

    _main('foo', is_debug=is_debug)

    if is_debug:
        mock_logging.basicConfig.assert_called_once_with(level=mock_logging.DEBUG)
    else:
        mock_logging.basicConfig.assert_not_called()




from typing import List, Iterator
import os
import logging

from ..module import SafeFilenameModule


logger = logging.getLogger(__name__)


class PackageScanner:
    """
    Scans a package for all the Python modules within it.

    Usage:
        package = SafeFilenameModule('mypackage', '/path/to/mypackage/__init__.py')
        scanner = PackageScanner(package)
        modules = scanner.scan_for_modules()
    """
    def __init__(self, package: SafeFilenameModule) -> None:
        self.package = package

    def scan_for_modules(self) -> List[SafeFilenameModule]:
        """
        Returns:
            List of module names (list(str, ...)).
        """
        package_directory = os.path.dirname(self.package.filename)
        modules: List[SafeFilenameModule] = []

        for module_filename in self._get_python_files_inside_package(package_directory):
            module_name = self._module_name_from_filename(module_filename, package_directory)
            modules.append(
                SafeFilenameModule(module_name, module_filename)
            )

        return modules

    def _get_python_files_inside_package(self, directory: str) -> Iterator[str]:
        """
        Get a list of Python files within the supplied package directory.
         Return:
            Generator of Python file names.
        """
        for root, dirs, files in os.walk(directory):
            # Don't include directories that aren't Python packages,
            # nor their subdirectories.
            if '__init__.py' not in files:
                for d in list(dirs):
                    dirs.remove(d)
                continue

            # Don't include hidden directories.
            dirs_to_remove = [d for d in dirs if self._should_ignore_dir(d)]
            for d in dirs_to_remove:
                dirs.remove(d)

            for filename in files:
                if self._is_python_file(filename):
                    yield os.path.join(root, filename)

    def _should_ignore_dir(self, directory: str) -> bool:
        # TODO: make this configurable.
        # Skip adding directories that are hidden, or look like Django migrations.
        return directory.startswith('.') or directory == 'migrations'

    def _is_python_file(self, filename: str) -> bool:
        """
        Given a filename, return whether it's a Python file.

        Args:
            filename (str): the filename, excluding the path.
        Returns:
            bool: whether it's a Python file.
        """
        return not filename.startswith('.') and filename.endswith('.py')

    def _module_name_from_filename(self, filename_and_path: str, package_directory: str) -> str:
        """
        Args:
            filename_and_path (string) - the full name of the Python file.
            package_directory (string) - the full path of the top level Python package directory.
         Returns:
            Absolute module name for importing (string).
        """
        container_directory, package_name = os.path.split(package_directory)
        internal_filename_and_path = filename_and_path[len(package_directory):]
        internal_filename_and_path_without_extension = internal_filename_and_path[1:-3]
        components = [package_name] + internal_filename_and_path_without_extension.split('/')
        if components[-1] == '__init__':
            components.pop()
        return '.'.join(components)



import os
import sys

import pytest

from layer_linter.contract import get_contracts, Layer
from layer_linter.dependencies import ImportPath
from layer_linter.module import Module


class TestGetContracts:
    def test_happy_path(self):
        self._initialize_test()

        contracts = get_contracts(self.filename_and_path, package_name='singlecontractfile')

        assert len(contracts) == 2
        expected_contracts = [
            {
                'name': 'Contract A',
                'packages': ['singlecontractfile.foo', 'singlecontractfile.bar'],
                'layers': ['one', 'two'],
            },
            {
                'name': 'Contract B',
                'packages': ['singlecontractfile'],
                'layers': ['one', 'two', 'three'],
                'whitelisted_paths': [
                    ('baz.utils', 'baz.three.green'),
                    ('baz.three.blue', 'baz.two'),
                ],
            },
        ]
        sorted_contracts = sorted(contracts, key=lambda i: i.name)
        for contract_index, contract in enumerate(sorted_contracts):
            expected_data = expected_contracts[contract_index]
            assert contract.name == expected_data['name']

            for package_index, package in enumerate(contract.containers):
                expected_package_name = expected_data['packages'][package_index]
                assert package == Module(expected_package_name)

            for layer_index, layer in enumerate(contract.layers):
                expected_layer_data = expected_data['layers'][layer_index]
                assert isinstance(layer, Layer)
                assert layer.name == expected_layer_data

            for whitelisted_index, whitelisted_path in enumerate(contract.whitelisted_paths):
                expected_importer, expected_imported = expected_data['whitelisted_paths'][
                    whitelisted_index]
                assert isinstance(whitelisted_path, ImportPath)
                assert whitelisted_path.importer == Module(expected_importer)
                assert whitelisted_path.imported == Module(expected_imported)

    def test_container_does_not_exist(self):
        self._initialize_test('layers_with_missing_container.yml')

        with pytest.raises(ValueError) as e:
            get_contracts(self.filename_and_path, package_name='singlecontractfile')

        assert str(e.value) == "Invalid container 'singlecontractfile.missing': no such package."

    def _initialize_test(self, config_filename='layers.yml'):
        # Append the package directory to the path.
        dirname = os.path.dirname(__file__)
        package_dirname = os.path.join(dirname, '..', 'assets', 'singlecontractfile')
        sys.path.append(package_dirname)

        # Set the full config filename and path as an instance attribute.
        self.filename_and_path = os.path.join(package_dirname, config_filename)



1 / 0  # Check that Runtime errors don't prevent this from being analyzed.

from differentimporttypes.one import alpha  # Import from: absolute import of module.
from differentimporttypes import one  # Import from: absolute import of package.

from . import beta  # Import from: relative, same level, module.
from . import gamma  # Import from: relative, same level, package

from .gamma import foo  # Import from: relative, down a level, module.

from ..two import alpha  # Import from: relative, up a level, module.

from .delta import some_function  # Import from: a function. NOT SUPPORTED YET.

def wrapper():
    from . import epsilon  # Import from: inside a function.


import differentimporttypes.three  # Import: absolute import of package.
import differentimporttypes.four.alpha  # Import: absolute import of module.

# Check that imports from other packages aren't included.
import pytest
from pytest import mark



from typing import List
import argparse
import os
import sys
import logging
import importlib

from .module import SafeFilenameModule
from .dependencies import DependencyGraph
from .contract import get_contracts, Contract, ContractParseError
from .report import (
    get_report_class, ConsolePrinter, VERBOSITY_QUIET, VERBOSITY_NORMAL, VERBOSITY_HIGH)


logger = logging.getLogger(__name__)

EXIT_STATUS_SUCCESS = 0
EXIT_STATUS_ERROR = 1

def create_parser():
    parser = argparse.ArgumentParser(
        description='Checks that your project follows a custom-defined '
        'layered architecture, based on a layers.yml file.'
    )

    parser.add_argument(
        'package_name',
        help='The name of the Python package to validate.'
    )

    parser.add_argument(
        '--config',
        required=False,
        help="The YAML file describing your contract(s). If not supplied, Layer Linter will "
             "look for a file called 'layers.yml' inside the current directory.",

    )

    parser.add_argument(
        '-v',
        '--verbose',
        # Using a count allows us to increase verbosity levels using -vv, -vvv etc., should they
        # be necessary in future.
        action='count',
        default=0,
        dest='verbosity_count',
        help="Increase verbosity."
    )

    parser.add_argument(
        '--quiet',
        required=False,
        action='store_true',
        dest='is_quiet',
        help="Do not output anything on success."
    )

    parser.add_argument(
        '--debug',
        required=False,
        action="store_true",
        dest='is_debug',
        help="Whether to display debug information.",
    )

    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()
    return _main(
        package_name=args.package_name,
        config_filename=args.config,
        is_debug=args.is_debug,
        verbosity_count=args.verbosity_count,
        is_quiet=args.is_quiet)


def _main(package_name, config_filename=None, is_debug=False,
          verbosity_count=0, is_quiet=False):

    if is_debug:
        logging.basicConfig(level=logging.DEBUG)

    try:
        package = _get_package(package_name)
    except ValueError as e:
        _print_package_name_error_and_help(str(e))
        return EXIT_STATUS_ERROR

    try:
        contracts = _get_contracts(config_filename, package_name)
    except Exception as e:
        ConsolePrinter.print_error(str(e))
        return EXIT_STATUS_ERROR

    graph = DependencyGraph(package=package)

    try:
        verbosity = _normalise_verbosity(verbosity_count, is_quiet)
    except Exception as e:
        ConsolePrinter.print_error(str(e))
        return EXIT_STATUS_ERROR

    report_class = get_report_class(verbosity)
    report = report_class(graph)

    for contract in contracts:
        try:
            contract.check_dependencies(graph)
        except Exception as e:
            ConsolePrinter.print_error(str(e))
            return EXIT_STATUS_ERROR
        report.add_contract(contract)

    report.output()

    if report.has_broken_contracts:
        return EXIT_STATUS_ERROR
    else:
        return EXIT_STATUS_SUCCESS


def _normalise_verbosity(verbosity_count: int, is_quiet: bool) -> int:
    """
    Validate verbosity, and parse quiet mode into a verbosity level.

    Args:
        verbosity_count (int): the number of 'v's passed as command line arguments. For example,
                               -vv would be 2.
        is_quiet (bool):       whether the '--quiet' flag was passed.

    Returns:
        Verbosity level (int): either VERBOSITY_QUIET, VERBOSITY_NORMAL or VERBOSITY_HIGH.
    """
    VERBOSITY_BY_COUNT = (VERBOSITY_NORMAL, VERBOSITY_HIGH)

    if is_quiet:
        if verbosity_count > 0:
            raise RuntimeError(
                "Invalid parameters: quiet and verbose called together. Choose one or the other.")
        return VERBOSITY_QUIET

    try:
        return VERBOSITY_BY_COUNT[verbosity_count]
    except IndexError:
        raise RuntimeError(
            "That level of verbosity is not supported. "
            "Maximum verbosity is -{}.".format('v' * (len(VERBOSITY_BY_COUNT) - 1)))


def _get_package(package_name: str) -> SafeFilenameModule:
    """
    Get the package as a SafeFilenameModule.

    Raises ValueError, with appropriate user-facing message, if the package name is
    not valid.
    """
    if '.' in package_name:
        raise ValueError("Package name must be the root name, no '.' allowed.")
    if ('/' in package_name) or ('\\' in package_name):
        raise ValueError("The package name should not be a directory, it should be the name of "
                         "the importable Python package.")

    # Add current directory to the path, as this doesn't happen automatically.
    sys.path.insert(0, os.getcwd())

    # Attempt to locate the package file.
    package_filename = importlib.util.find_spec(package_name)
    if not package_filename:
        logger.debug("sys.path: {}".format(sys.path))
        raise ValueError("Could not find package '{}' in your Python path.".format(package_name))
    assert package_filename.origin  # For type checker.
    return SafeFilenameModule(name=package_name, filename=package_filename.origin)


def _get_contracts(config_filename: str, package_name: str) -> List[Contract]:
    # Parse contracts file.
    if config_filename is None:
        config_filename = os.path.join(os.getcwd(), 'layers.yml')
    try:
        return get_contracts(filename=config_filename, package_name=package_name)
    except FileNotFoundError as e:
        raise RuntimeError("{}: {}".format(e.strerror, e.filename))
    except ContractParseError as e:
        raise RuntimeError('Error parsing contract: {}'.format(e))


def _print_package_name_error_and_help(error_text):
    ConsolePrinter.print_heading('Invalid package name',
                                 ConsolePrinter.HEADING_LEVEL_TWO,
                                 ConsolePrinter.ERROR)
    ConsolePrinter.print_error(error_text)
    ConsolePrinter.new_line()
    ConsolePrinter.print_heading('Tip', ConsolePrinter.HEADING_LEVEL_THREE,
                                 ConsolePrinter.ERROR)
    ConsolePrinter.print_error('Your package should either be in the current working directory, ')
    ConsolePrinter.print_error('or installed (e.g. in your virtual environment).')
    ConsolePrinter.new_line()
    ConsolePrinter.print_error('If your package has a setup.py, the easiest way to install it is ')
    ConsolePrinter.print_error('to run the following command, which installs it, while keeping ')
    ConsolePrinter.print_error('it editable:')
    ConsolePrinter.new_line()
    ConsolePrinter.indent_cursor()
    ConsolePrinter.print_error('pip install -e path/to/setup.py')
    ConsolePrinter.new_line()
    ConsolePrinter.print_error("Alternatively, you may run Layer Linter from the directory that ")
    ConsolePrinter.print_error("contains your package directory. If your layers.yml is located ")
    ConsolePrinter.print_error("elsewhere, you may specify its file path using the config flag. ")
    ConsolePrinter.print_error("For example:")
    ConsolePrinter.new_line()
    ConsolePrinter.indent_cursor()
    ConsolePrinter.print_error(
        'layer-lint mypackage --config path/to/layers.yml')
    ConsolePrinter.new_line()



from unittest import mock
import importlib
import pytest
from layer_linter import contract
from layer_linter.module import Module
from layer_linter.contract import Contract, Layer
import logging
import sys

logger = logging.getLogger('layer_linter')
logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)


class StubDependencyGraph:
    """
    Stub of the DependencyGraph for use in tests.

    Args:
        descendants: Dictionary keyed with Modules, whose values are the descendants of
                     those modules. This should not include the module itself.
        paths:       Nested dictionary, in the form:

                     {
                        upstream_module: {
                            downstream_module: [downstream_module, ..., upstream_module],
                            ...
                        }
                        ...
                     }
                    A call to .find_path(downstream=downstream_module, upstream=upstream_module)
                    will return the supplied path, if it is present in the dictionary. Otherwise,
                    the stub will return None.
        modules:    List of all modules in the graph.
    Usage:

        graph = StubDependencyGraph(
            descendants={
                Module('foo.one'): [
                    Module('foo.one.red'), Module('foo.one.green')
                ],
            },
            paths = {
                Module('foo.one.green'): {
                    Module('foo.three.blue.alpha'): [
                        Module('foo.one.green'), Module('baz'), Module('foo.three.blue.alpha')
                    ],
                },
                ...
            },
            modules [Module('foo.one'), Module('foo.one.red'), Module('foo.one.green')],
        )

    """
    def __init__(self, descendants=None, paths=None, modules=None):
        self.descendants = descendants if descendants else {}
        self.paths = paths if paths else {}
        self.modules = modules if modules else []

    def get_descendants(self, module):
        try:
            return self.descendants[module]
        except KeyError:
            return []

    def find_path(self, downstream, upstream, ignore_paths=None):
        try:
            return self.paths[upstream][downstream]
        except KeyError:
            return None

    def __contains__(self, item):
        return item in self.modules


class TestContractCheck:
    def test_kept_contract(self):
        contract = Contract(
            name='Foo contract',
            containers=(
                Module('foo.blue'),
                Module('foo.green'),
            ),
            layers=(
                Layer('three'),
                Layer('two'),
                Layer('one'),
            ),
            whitelisted_paths=mock.sentinel.whitelisted_paths,
        )
        graph = StubDependencyGraph(
            descendants={
                Module('foo.green.one'): [
                    Module('foo.green.one.alpha'),
                    Module('foo.green.one.beta'),
                ],
                Module('foo.green.three'): [
                    Module('foo.green.three.alpha'),
                    Module('foo.green.three.beta'),
                ],
            },
            paths={
                # Include some allowed paths.
                Module('foo.blue.two'): {
                    # Layer directly importing a layer below it.
                    Module('foo.blue.three'): [Module('foo.blue.three'), Module('foo.blue.two')],
                },
                Module('foo.blue.one'): {
                    # Layer directly importing two layers below.
                    Module('foo.blue.three'): [Module('foo.blue.three'), Module('foo.blue.one')],
                },
                Module('foo.green.three'): {
                    # Layer importing higher up layer, but from another container.
                    Module('foo.blue.one'): [Module('foo.blue.one'), Module('foo.green.three')],
                },
                Module('foo.green.three.beta'): {
                    # Module inside layer importing another module in same layer.
                    Module('foo.green.three.alpha'): [Module('foo.green.three.alpha'),
                                                      Module('foo.green.three.beta')],
                },
                Module('foo.green.one.alpha'): {
                    # Module inside layer importing a module inside a lower layer.
                    Module('foo.green.three.alpha'): [Module('foo.green.three.alpha'),
                                                      Module('foo.green.one.alpha')]
                },
            },
            modules=[
                Module('foo.green'),
                Module('foo.green.one'),
                Module('foo.green.one.alpha'),
                Module('foo.green.one.beta'),
                Module('foo.green.two'),
                Module('foo.green.three'),
                Module('foo.blue'),
                Module('foo.blue.one'),
                Module('foo.blue.two'),
                Module('foo.blue.three'),
            ]
        )

        contract.check_dependencies(graph)

        assert contract.is_kept is True

    def test_broken_contract(self):
        contract = Contract(
            name='Foo contract',
            containers=(
                Module('foo.blue'),
                Module('foo.green'),
            ),
            layers=(
                Layer('three'),
                Layer('two'),
                Layer('one'),
            ),
        )
        graph = StubDependencyGraph(
            descendants={
                Module('foo.green.one'): [
                    Module('foo.green.one.alpha'),
                    Module('foo.green.one.beta'),
                ],
                Module('foo.green.three'): [
                    Module('foo.green.three.alpha'),
                    Module('foo.green.three.beta'),
                ],
            },
            paths={
                Module('foo.blue.two'): {
                    # An allowed path: layer directly importing a layer below it.
                    Module('foo.blue.three'): [Module('foo.blue.three'), Module('foo.blue.two')],
                    # Disallowed path: layer directly importing a layer above it.
                    Module('foo.blue.one'): [Module('foo.blue.one'), Module('foo.blue.two')],
                },
                Module('foo.green.three.alpha'): {
                    # Module inside layer importing a module inside a higher layer.
                    Module('foo.green.one.alpha'): [Module('foo.green.one.alpha'),
                                                    Module('foo.green.three.alpha')],
                },
            },
            modules=[
                Module('foo.green'),
                Module('foo.green.one'),
                Module('foo.green.one.alpha'),
                Module('foo.green.one.beta'),
                Module('foo.green.two'),
                Module('foo.green.three'),
                Module('foo.blue'),
                Module('foo.blue.one'),
                Module('foo.blue.two'),
                Module('foo.blue.three'),
            ]
        )

        contract.check_dependencies(graph)

        assert contract.is_kept is False
        assert contract.illegal_dependencies == [
            [Module('foo.blue.one'), Module('foo.blue.two')],
            [Module('foo.green.one.alpha'), Module('foo.green.three.alpha')]
        ]

    def test_unchecked_contract_raises_exception(self):
        contract = Contract(
            name='Foo contract',
            containers=(
                'foo',
            ),
            layers=(
                Layer('three'),
                Layer('two'),
                Layer('one'),
            ),
        )

        with pytest.raises(RuntimeError) as excinfo:
            contract.is_kept
        assert 'Cannot check whether contract is ' \
            'kept until check_dependencies is called.' in str(excinfo.value)

    def test_broken_contract_via_other_layer(self):
        # If an illegal import happens via another layer, we don't want to report it
        # (as it will already be reported).

        contract = Contract(
            name='Foo contract',
            containers=(
                'foo',
            ),
            layers=(
                Layer('three'),
                Layer('two'),
                Layer('one'),
            ),
        )
        graph = StubDependencyGraph(
            descendants={},
            paths={
                Module('foo.three'): {
                    Module('foo.two'): [Module('foo.two'), Module('foo.three')],
                    Module('foo.one'): [Module('foo.one'), Module('foo.two'), Module('foo.three')],
                },
                Module('foo.two'): {
                    Module('foo.one'): [Module('foo.one'), Module('foo.two')],
                },
            },
            modules=[
                Module('foo.one'),
                Module('foo.two'),
                Module('foo.three'),
            ]
        )

        contract.check_dependencies(graph)

        assert contract.illegal_dependencies == [
            [Module('foo.one'), Module('foo.two')],
            [Module('foo.two'), Module('foo.three')],
        ]

    

    @pytest.mark.parametrize('longer_first', (True, False))
    def test_only_shortest_violation_is_reported(self, longer_first):
        contract = Contract(
            name='Foo contract',
            containers=(
                'foo',
            ),
            layers=(
                Layer('two'),
                Layer('one'),
            ),
        )

        # These are both dependency violations, but it's more useful just to report
        # the more direct violation.
        if longer_first:
            paths = {
                Module('foo.two'): {
                    Module('foo.one.alpha'): [
                        Module('foo.one.alpha'), Module('foo.one.alpha.green'),
                        Module('foo.another'), Module('foo.two'),
                    ],
                    Module('foo.one.alpha.green'): [
                        Module('foo.one.alpha.green'), Module('foo.another'), Module('foo.two'),
                    ],

                },
            }
        else:
            paths = {
                Module('foo.two'): {
                    Module('foo.one.alpha'): [
                        Module('foo.one.alpha'), Module('foo.another'), Module('foo.two'),
                    ],
                    Module('foo.one.alpha.green'): [
                        Module('foo.one.alpha.green'), Module('foo.one.alpha'),
                        Module('foo.another'), Module('foo.two'),
                    ],

                },
            }
        graph = StubDependencyGraph(
            descendants={
                Module('foo.one'): [Module('foo.one.alpha'), Module('foo.one.beta'),
                                    Module('foo.one.alpha.blue'), Module('foo.one.alpha.green')],
            },
            paths=paths,
            modules=[Module('foo.one'), Module('foo.two')]
        )

        contract.check_dependencies(graph)

        if longer_first:
            assert contract.illegal_dependencies == [
                [Module('foo.one.alpha.green'), Module('foo.another'), Module('foo.two')],
            ]
        else:
            assert contract.illegal_dependencies == [
                [Module('foo.one.alpha'), Module('foo.another'), Module('foo.two')],
            ]

    @pytest.mark.parametrize(
        'is_optional', (True, False),
    )
    def test_missing_contract(self, is_optional):
        contract = Contract(
            name='Foo contract',
            containers=(
                'foo.one',
                'foo.two',
            ),
            layers=(
                Layer('blue'),
                Layer('yellow', is_optional=is_optional),  # Missing from foo.two.
                Layer('green'),
            ),
        )
        graph = StubDependencyGraph(
            modules=[
                Module('foo.one'),
                Module('foo.one.blue'),
                Module('foo.one.blue.alpha'),
                Module('foo.one.yellow'),
                Module('foo.one.green'),
                Module('foo.two'),
                Module('foo.two.blue'),
                Module('foo.two.blue.alpha'),
                Module('foo.two.green'),
            ]
        )

        if is_optional:
            # Should pass.
            contract.check_dependencies(graph)
        else:
            with pytest.raises(ValueError) as e:
                contract.check_dependencies(graph)

            assert str(e.value) == (
                "Missing layer in container 'foo.two': module foo.two.yellow does not exist."
            )


@mock.patch.object(importlib.util, 'find_spec')
class TestContractFromYAML:

    def test_parentheses_indicate_optional_layer(self, mock_find_spec):
        data = {
            'containers': ['mypackage.foo'],
            'layers': ['one', '(two)'],
        }

        parsed_contract = contract.contract_from_yaml('Contract Foo', data, 'mypackage')

        assert not parsed_contract.layers[0].is_optional
        assert parsed_contract.layers[1].is_optional

    def test_incorrect_whitelisted_path_format(self, mock_find_spec):
        data = {
            'containers': ['mypackage.foo', 'mypackage.bar'],
            'layers': ['one', 'two'],
            'whitelisted_paths': [
                'not the right format',
            ]
        }

        with pytest.raises(ValueError) as exception:
            contract.contract_from_yaml('Contract Foo', data, 'mypackage')
        assert str(exception.value) == (
            'Whitelisted paths must be in the format '
            '"importer.module <- imported.module".'
        )

    def test_container_not_in_package(self, mock_find_spec):
        data = {
            'containers': ['mypackage.foo', 'anotherpackage.foo'],
            'layers': ['one', 'two'],
        }

        with pytest.raises(ValueError) as exception:
            contract.contract_from_yaml('Contract Foo', data, 'mypackage')
        assert str(exception.value) == (
            "Invalid container 'anotherpackage.foo': containers must be either a subpackage of "
            "'mypackage', or 'mypackage' itself."
        )



import os
import sys

from layer_linter.cmdline import _main, EXIT_STATUS_SUCCESS, EXIT_STATUS_ERROR

import pytest

assets_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'assets'))


@pytest.mark.parametrize(
    'verbosity_count, is_quiet, should_always_fail',
    (
        (0, False, False),  # no args
        (0, True, False),  # --quiet
        (1, False, False),  # -v
        (1, True, True),  # -v --quiet: should always fail.
    )
)
class TestMain:
    def test_success(self, verbosity_count, is_quiet, should_always_fail):
        package_name = 'successpackage'
        self._chdir_and_add_to_system_path(package_name)

        result = _main(package_name, verbosity_count=verbosity_count, is_quiet=is_quiet)

        if should_always_fail:
            assert result == EXIT_STATUS_ERROR
        else:
            assert result == EXIT_STATUS_SUCCESS

    def test_failure(self, verbosity_count, is_quiet, should_always_fail):
        package_name = 'failurepackage'
        self._chdir_and_add_to_system_path(package_name)

        result = _main(package_name, verbosity_count=verbosity_count, is_quiet=is_quiet)

        assert result == EXIT_STATUS_ERROR

    def test_specify_config_file(self, verbosity_count, is_quiet, should_always_fail):
        package_name = 'dependenciespackage'
        self._chdir_and_add_to_system_path(package_name)

        result = _main(
            package_name,
            config_filename='../successpackage/layers_alternative.yml',
            verbosity_count=verbosity_count,
            is_quiet=is_quiet)

        if should_always_fail:
            assert result == EXIT_STATUS_ERROR
        else:
            assert result == EXIT_STATUS_SUCCESS

    def test_missing_container(self, verbosity_count, is_quiet, should_always_fail):
        package_name = 'successpackage'
        self._chdir_and_add_to_system_path(package_name)

        result = _main(package_name,
                       config_filename='layers_with_missing_container.yml',
                       verbosity_count=verbosity_count, is_quiet=is_quiet)

        assert result == EXIT_STATUS_ERROR

    def _chdir_and_add_to_system_path(self, package_name):
        path = os.path.join(assets_path, package_name)
        sys.path.append(path)
        os.chdir(path)



from typing import Tuple, List
import os
import sys

import pytest

from layer_linter.dependencies.analysis import DependencyAnalyzer
from layer_linter.dependencies.path import ImportPath
from layer_linter.module import Module, SafeFilenameModule


class TestDependencyAnalyzer:
    def test_success(self):
        package = Module('analyzerpackage')
        modules = self._build_modules(
            package_name=package.name,
            tuples=(
                ('analyzerpackage', '__init__.py'),
                ('analyzerpackage.utils', 'utils.py'),
                ('analyzerpackage.one', 'one/__init__.py'),
                ('analyzerpackage.one.alpha', 'one/alpha.py'),
                ('analyzerpackage.one.beta', 'one/beta.py'),
                ('analyzerpackage.one.gamma', 'one/gamma.py'),
                ('analyzerpackage.two', 'two/__init__.py'),
                ('analyzerpackage.two.alpha', 'two/alpha.py'),
                ('analyzerpackage.two.beta', 'two/beta.py'),
                ('analyzerpackage.two.gamma', 'two/gamma.py'),
            ),
        )

        analyzer = DependencyAnalyzer(modules, package)
        import_paths = analyzer.determine_import_paths()

        # N.B. We expect only the leaf imports, it's just noise to have parent packages
        # in the graph too. However, if there is a separate parent import, that should be included.
        expected_import_paths = self._build_import_paths(
            tuples=(
                ('analyzerpackage.utils', 'analyzerpackage.one'),
                ('analyzerpackage.utils', 'analyzerpackage.two.alpha'),
                ('analyzerpackage.one.beta', 'analyzerpackage.one.alpha'),
                ('analyzerpackage.one.gamma', 'analyzerpackage.one.beta'),
                ('analyzerpackage.two.alpha', 'analyzerpackage.one.alpha'),
                ('analyzerpackage.two.beta', 'analyzerpackage.one.alpha'),
                ('analyzerpackage.two.gamma', 'analyzerpackage.two.beta'),
                ('analyzerpackage.two.gamma', 'analyzerpackage.utils'),
            )
        )

        assert set(import_paths) == set(expected_import_paths)

    def test_all_different_import_types(self):
        """
        Test a single file with lots of different import types. Note that all of the other
        modules are empty files.
        """
        package = Module('differentimporttypes')
        modules = self._build_modules(
            package_name=package.name,
            tuples=(
                # The first module is the one with the code we'll analyze.
                ('differentimporttypes.one.importer', 'one/importer.py'),
                ('differentimporttypes', '__init__.py'),
                ('differentimporttypes.one', 'one/__init__.py'),
                ('differentimporttypes.one.alpha', 'one/alpha.py'),
                ('differentimporttypes.one.beta', 'one/beta.py'),
                ('differentimporttypes.one.gamma', 'one/gamma/__init__.py'),
                ('differentimporttypes.one.gamma.foo', 'one/gamma/foo.py'),
                ('differentimporttypes.one.delta', 'one/delta.py'),
                ('differentimporttypes.one.epsilon', 'one/epsilon.py'),

                ('differentimporttypes.two', 'two/__init__.py'),
                ('differentimporttypes.two.alpha', 'two/alpha.py'),
                ('differentimporttypes.three', 'three.py'),
                ('differentimporttypes.four', 'four/__init__.py'),
                ('differentimporttypes.four.alpha', 'four/alpha.py'),
                # Some other modules, not imported.
                ('differentimporttypes.five', 'five/__init__.py'),
                ('differentimporttypes.five.alpha', 'five/alpha.py'),
            ),
        )

        analyzer = DependencyAnalyzer(modules, package)
        import_paths = analyzer.determine_import_paths()

        expected_import_paths = self._build_import_paths(
            tuples=(
                ('differentimporttypes.one.importer', 'differentimporttypes.one'),
                ('differentimporttypes.one.importer', 'differentimporttypes.one.alpha'),
                ('differentimporttypes.one.importer', 'differentimporttypes.one.beta'),
                ('differentimporttypes.one.importer', 'differentimporttypes.one.gamma'),
                ('differentimporttypes.one.importer', 'differentimporttypes.one.gamma.foo'),
                ('differentimporttypes.one.importer', 'differentimporttypes.two.alpha'),
                ('differentimporttypes.one.importer', 'differentimporttypes.one.delta'),
                ('differentimporttypes.one.importer', 'differentimporttypes.one.epsilon'),
                ('differentimporttypes.one.importer', 'differentimporttypes.three'),
                ('differentimporttypes.one.importer', 'differentimporttypes.four.alpha'),
            )
        )

        assert set(import_paths) == set(expected_import_paths)

    def _build_modules(self,
                       package_name: str,
                       tuples: Tuple[Tuple[str, str]]) -> List[SafeFilenameModule]:
        package_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                '..', '..', 'assets', package_name, package_name,
            )
        )
        modules = []
        for module_name, module_filename in tuples:
            modules.append(
                SafeFilenameModule(
                    name=module_name,
                    filename=os.path.join(package_path, module_filename),
                )
            )
        return modules

    def _build_import_paths(self, tuples: Tuple[Tuple[str, str]]) -> List[ImportPath]:
        import_paths = []
        for importer, imported in tuples:
            import_paths.append(
                ImportPath(importer=Module(importer), imported=Module(imported))
            )
        return import_paths



import logging
from typing import List, Optional, Any
import networkx  # type: ignore
from networkx.algorithms import shortest_path  # type: ignore

from ..module import Module, SafeFilenameModule
from .path import ImportPath
from .scanner import PackageScanner
from .analysis import DependencyAnalyzer


logger = logging.getLogger(__name__)


class DependencyGraph:
    """
    A graph of the internal dependencies in a Python package.

    Usage:
        graph = DependencyGraph(
            SafeFilenameModule('mypackage', '/path/to/mypackage/__init__.py')
        )

        path = graph.find_path(
            downstream=Module('mypackage.foo.one'),
            upstream=Module('mypackage.bar.two'),
            ignore_paths=[
                ImportPath(
                    importer=Module('mypackage.foo'),
                    imported=Module('mypackage.baz.three'),
                ),
            ]
        )

        descendants = graph.get_descendants(Module('mypackage.foo'))
    """
    def __init__(self, package: SafeFilenameModule) -> None:
        scanner = PackageScanner(package)
        self.modules = scanner.scan_for_modules()

        self._networkx_graph = networkx.DiGraph()
        self.dependency_count = 0

        analyzer = DependencyAnalyzer(modules=self.modules, package=package)
        for import_path in analyzer.determine_import_paths():
            self._add_path_to_networkx_graph(import_path)
            self.dependency_count += 1

        self.module_count = len(self.modules)

    def get_modules_directly_imported_by(self, importer: Module) -> List[Module]:
        """
        Returns all the modules directly imported by the importer.
        """
        if importer in self._networkx_graph:
            return self._networkx_graph.successors(importer)
        else:
            # Nodes that do not import anything are not present in the networkx graph.
            return []

    def find_path(self,
                  downstream: Module, upstream: Module,
                  ignore_paths: Optional[List[ImportPath]] = None) -> List[Module]:
        """
        Find a list of module names showing the dependency path between the downstream and the
        upstream module, or None if there is no dependency.

        For example, given an upstream module a and a downstream module d:

                - [d, a] will be returned if d directly imports a.
                - [d, c, b, a] will be returned if d imports c, which imports b, which imports a.
                - None will be returned if d does not import a (even indirectly).
        """
        logger.debug("Finding path from '{}' up to '{}'.".format(downstream, upstream))
        ignore_paths = ignore_paths if ignore_paths else []

        removed_paths = self._remove_paths_from_networkx_graph(ignore_paths)

        try:
            path = shortest_path(self._networkx_graph, downstream, upstream)
        except (networkx.NetworkXNoPath, networkx.exception.NodeNotFound):
            # Either there is no path, or one of the modules doesn't even exist.
            path = None
        else:
            path = tuple(path)

        self._restore_paths_to_networkx_graph(removed_paths)

        return path

    def get_descendants(self, module: Module) -> List[Module]:
        """
        Returns:
            List of modules that are within the supplied module.
        """
        descendants: List[Module] = []
        for candidate in self.modules:
            if candidate.name.startswith('{}.'.format(module.name)):
                descendants.append(candidate)
        return descendants

    def _add_path_to_networkx_graph(self, import_path: ImportPath) -> None:
        self._networkx_graph.add_edge(import_path.importer, import_path.imported)

    def _remove_path_from_networkx_graph(self, import_path: ImportPath) -> None:
        self._networkx_graph.remove_edge(import_path.importer, import_path.imported)

    def _import_path_is_in_networkx_graph(self, import_path: ImportPath) -> bool:
        return self._networkx_graph.has_successor(
            import_path.importer, import_path.imported
        )

    def _remove_paths_from_networkx_graph(
        self, import_paths: List[ImportPath]
    ) -> List[ImportPath]:
        """
        Given a list of ImportPaths, remove any that exist from the graph.

        Returns:
             List of removed ImportPaths.
        """
        removed_paths = []
        for import_path in import_paths:
            if self._import_path_is_in_networkx_graph(import_path):
                self._remove_path_from_networkx_graph(import_path)
                removed_paths.append(import_path)
        return removed_paths

    def _restore_paths_to_networkx_graph(self, import_paths: List[ImportPath]) -> None:
        for import_path in import_paths:
            self._add_path_to_networkx_graph(import_path)

    def __contains__(self, item: Any) -> bool:
        """
        Returns whether or not the given item is a Module in the dependency graph.

        Usage:
            if module in graph:
                ...
        """
        return item in self.modules



from layer_linter.dependencies.path import ImportPath
from layer_linter.module import Module


class TestImportPath:
    def test_repr(self):
        import_path = ImportPath(
            importer=Module('foo'), imported=Module('bar')
        )
        assert repr(import_path) == '<ImportPath: foo <- bar>'

    def test_equals(self):
        a = ImportPath(importer=Module('foo'), imported=Module('bar'))
        b = ImportPath(importer=Module('foo'), imported=Module('bar'))
        c = ImportPath(importer=Module('foo'), imported=Module('foo.baz'))

        assert a == b
        assert a != c
        # Also non-ImportPath instances should not be treated as equal.
        assert a != 'foo'


    def test_hash(self):
        a = ImportPath(importer=Module('foo'), imported=Module('bar'))
        b = ImportPath(importer=Module('foo'), imported=Module('bar'))
        c = ImportPath(importer=Module('bar'), imported=Module('foo'))

        assert hash(a) == hash(b)
        assert hash(a) != hash(c)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'networkx~=2.2',
    'PyYAML~=4.2b1',
    'click~=6.7',
]

setup(
    author="David Seddon",
    author_email='david@seddonym.me',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="Layer Linter checks that your project follows a custom-defined layered architecture.",
    install_requires=requirements,
    license="BSD license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='layer-linter layer-lint',
    name='layer-linter',
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    url='https://github.com/seddonym/layer_linter',
    version='0.11.1',
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'layer-lint = layer_linter.cmdline:main',
        ],
    },
)



from unittest.mock import patch, Mock

from layer_linter.dependencies import graph as graph_module
from layer_linter.dependencies.path import ImportPath
from layer_linter.module import Module


class DependencyAnalyzerStub:
    def __init__(self, modules, package):
        self.import_paths = [
            ImportPath(importer=Module('foo.two'), imported=Module('foo.one')),
            ImportPath(importer=Module('foo.three'), imported=Module('foo.two')),
            ImportPath(importer=Module('foo.four'), imported=Module('foo.three')),
        ]

    def determine_import_paths(self):
        return self.import_paths


class PackageScannerStub:
    def __init__(self, package):
        self.modules = [
            Module('foo'),
            Module('foo.one'),
            Module('foo.one.alpha'),
            Module('foo.one.beta'),
            Module('foo.one.beta.green'),
            Module('foo.two'),
        ]

    def scan_for_modules(self):
        return self.modules


@patch.object(graph_module, 'DependencyAnalyzer', new=DependencyAnalyzerStub)
@patch.object(graph_module, 'PackageScanner', new=PackageScannerStub)
class TestDependencyGraph:
    PACKAGE = Mock(__name__='foo')

    def test_find_path_direct(self):
        graph = graph_module.DependencyGraph(self.PACKAGE)

        path = graph.find_path(upstream=Module('foo.one'), downstream=Module('foo.two'))

        assert path == (Module('foo.two'), Module('foo.one'))

    def test_find_path_indirect(self):
        graph = graph_module.DependencyGraph(self.PACKAGE)

        path = graph.find_path(upstream=Module('foo.one'), downstream=Module('foo.four'))

        assert path == (Module('foo.four'), Module('foo.three'), Module('foo.two'),
                        Module('foo.one'),)

    def test_find_path_nonexistent(self):
        graph = graph_module.DependencyGraph(self.PACKAGE)

        path = graph.find_path(upstream=Module('foo.four'), downstream=Module('foo.one'))

        assert path is None

    def test_direct_ignore_path_is_ignored(self):
        ignore_paths = (
            ImportPath(
                importer=Module('foo.two'), imported=Module('foo.one'),
            ),
        )
        graph = graph_module.DependencyGraph(self.PACKAGE)

        path = graph.find_path(
            upstream=Module('foo.one'), downstream=Module('foo.two'),
            ignore_paths=ignore_paths)

        assert path is None

    def test_indirect_ignore_path_is_ignored(self):
        ignore_paths = (
            ImportPath(
                importer=Module('foo.three'), imported=Module('foo.two'),
            ),
        )
        graph = graph_module.DependencyGraph(self.PACKAGE)

        path = graph.find_path(
            upstream=Module('foo.one'), downstream=Module('foo.four'),
            ignore_paths=ignore_paths)

        assert path is None

    def test_get_descendants_nested(self):

        graph = graph_module.DependencyGraph(self.PACKAGE)

        assert set(graph.get_descendants(Module('foo.one'))) == {
            Module('foo.one.alpha'),
            Module('foo.one.beta'),
            Module('foo.one.beta.green'),
        }

    def test_get_descendants_none(self):
        graph = graph_module.DependencyGraph(self.PACKAGE)

        assert graph.get_descendants(Module('foo.two')) == []

    def test_module_count(self):
        graph = graph_module.DependencyGraph(self.PACKAGE)

        # Assert the module count is the number of modules returned by
        # PackageScanner.scan_for_modules.
        assert graph.module_count == 6

    def test_dependency_count(self):
        graph = graph_module.DependencyGraph(self.PACKAGE)

        # Should be number of ImportPaths returned by DependencyAnalyzer.determine_import_paths.
        assert graph.dependency_count == 3

    def test_contains(self):
        graph = graph_module.DependencyGraph(self.PACKAGE)

        assert Module('foo.one.alpha') in graph
        assert Module('foo.one.omega') not in graph



from typing import Any

from ..module import Module


class ImportPath:
    """
    A direct dependency path between two modules. For example,
    if foo imports bar.baz, then the ImportPath is:
        ImportPath(
            importer=Module('foo'),
            imported=Module('bar.baz'),
        )
    """
    def __init__(self, importer: Module, imported: Module) -> None:
        self.importer = importer
        self.imported = imported

    def __str__(self) -> str:
        return "{} <- {}".format(self.importer, self.imported)

    def __repr__(self) -> str:
        return '<{}: {}>'.format(self.__class__.__name__, self)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ImportPath):
            return (self.importer, self.imported) == (other.importer, other.imported)
        else:
            return False

    def __hash__(self) -> int:
        return hash(str(self))
